/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */




/************************************************************************
 * thr.h: declaration of the Tcb class
 *
 * A Tcb represents a thread library thread.
************************************************************************/
#if !defined(_thread_src_thr_h_)
#define _thread_src_thr_h_


// constants related to the Tcb class
#define THR_NAMELEN				32
#define	DEFAULT_THR_STACK_SIZE	16384
#define THR_KEY_TABLE_SIZE		64



// the Tcb class
typedef struct Tcb* Thread;
struct Tcb : public Item
{
private:
	struct Stack
	{
		bool	used;
		char	mem[DEFAULT_THR_STACK_SIZE];

		Stack( void ) : used( false ) { }
	};

	static	Stack		stack_pool[];
	static	unsigned	next_stack;

	static	void*		GetStack( unsigned* stack_sizep );


public:
#if defined(ENABLE_PREEMPTION)

	typedef sigjmp_buf State;
	typedef int  (*Save_Func)(State, int);
	typedef void (*Rest_Func)(State, int);

#else /* defined(ENABLE_PREEMPTION) */

	typedef jmp_buf    State;
	typedef int  (*Save_Func)(State);
	typedef void (*Rest_Func)(State, int);

#endif /* defined(ENABLE_PREEMPTION) */
	typedef void* (*Func)(void *);


    struct {
        thread_t  tid;
        thread_t  ptid;
        int       errno_local;
        char      name[THR_NAMELEN+1];
    }             id;

    struct {
        unsigned  is_main   : 1;
        unsigned  own_st    : 1;
        unsigned  started   : 1;
        unsigned  exited    : 1;
        unsigned  killed    : 1;
        unsigned  suspended : 1;
        unsigned  blocked   : 1;
        unsigned  polling   : 1;
        unsigned  waiting   : 1;
        unsigned  reaped    : 1;
    }             flags;

    struct {
        State     state;
        void*     stack;
        void*     st_bot;
        void*     st_top;
        unsigned  stack_size;
    }             exec;

    struct {
        Func      func;
        void*     args;
        void*     result;
    }             call;

    struct {
        thread_t  waiter;
        thread_t* departed;
        void**    status;
    }             wait;

    struct {
        void*     key_data[THR_KEY_TABLE_SIZE];
    }             key;

    struct {
        Queue		mq;
		thread_t*	tidp;
		tag_t*		tagp;
        void*		buf;
        unsigned*	nitemsp;
    }				msg;

    struct thr_fd {
        WaitSet   waitset;
        thread_t  last_file;
		thread_t  last_socket;
    }             fd;

    struct {
        unsigned  nsigs;
    }             sig;


	static	Save_Func	save_context;
	static	Rest_Func	restore_context;

	static	Thread	Alloc( thread_t ptid,
						Tcb::Func func,
						void* args,
						unsigned flags,
						void* stack,
						unsigned stack_size,
						unsigned is_main );
	static	void	dummy_start( void );

	void	Init( thread_t tid,
					thread_t ptid,
					Tcb::Func func,
					void* args,
					unsigned flags,
					void* stack,
					unsigned stack_size,
					unsigned is_main );

			void	Bind( thread_t tid );
			void	Unbind( thread_t tid );
			bool	i_own( thread_t tid ) const;

#if defined(THREADLIB_DEBUG)
	virtual	void	WriteTo( ostream& s ) const;
#endif // defined(THREADLIB_DEBUG)
};


#endif /* !defined(_thread_src_thr_h_) */
